#include <cstdlib>
#include <cstring>
#include <cstdio>
#include "csutil.hxx"

#include "atypes.hxx"
#include "langnum.hxx"

#include "utf_info.cxx"
#define UTF_LST_LEN (sizeof(utf_lst) / (sizeof(unicode_info)))

#if !defined(_MSC_VER)
using namespace std;
#endif

/* only UTF-16 (BMP) implementation */
char * u16_u8(char * dest, int size, const w_char * src, int srclen) {
    char * u8 = dest;
    char * u8_max = u8 + size;
    const w_char * u2 = src;
    const w_char * u2_max = src + srclen;
    while ((u2 < u2_max) && (u8 < u8_max)) {
        if (u2->h) { // > 0xFF
            // XXX 4-byte haven't implemented yet.
            if (u2->h >= 0x08) {   // >= 0x800 (3-byte UTF-8 character)
                *u8 = 0xe0 + (u2->h >> 4);
                u8++;
                if (u8 < u8_max) {
                    *u8 = 0x80 + ((u2->h & 0xf) << 2) + (u2->l >> 6);
                    u8++;
                    if (u8 < u8_max) {
                        *u8 = 0x80 + (u2->l & 0x3f);
                        u8++;
                    }
                }
            } else { // < 0x800 (2-byte UTF-8 character)
                *u8 = 0xc0 + (u2->h << 2) + (u2->l >> 6);
                u8++;
                if (u8 < u8_max) {
                    *u8 = 0x80 + (u2->l & 0x3f);
                    u8++;
                }
            }
        } else { // <= 0xFF
            if (u2->l & 0x80) { // >0x80 (2-byte UTF-8 character)
                *u8 = 0xc0 + (u2->l >> 6);
                u8++;
                if (u8 < u8_max) {
                    *u8 = 0x80 + (u2->l & 0x3f);
                    u8++;
                }
            } else { // < 0x80 (1-byte UTF-8 character)
                *u8 = u2->l;
                u8++;
            }
        }
        u2++;
    }
    *u8 = '\0';
    return dest;
}


/* only UTF-16 (BMP) implementation */
int u8_u16(w_char * dest, int size, const char * src) {
    const char * u8 = src;
    w_char * u2 = dest;
    w_char * u2_max = u2 + size;
    
    while (*u8 && (u2 < u2_max)) {
    switch ((*u8) & 0xf0) {
        case 0x00:
        case 0x10:
        case 0x20:
        case 0x30:
        case 0x40:
        case 0x50:
        case 0x60:
        case 0x70: {
            u2->h = 0;
            u2->l = *u8;
            break;
        }
        case 0x80:
        case 0x90:
        case 0xa0:
        case 0xb0: {
            fprintf(stderr, "UTF-8 encoding error. Unexpected continuation bytes in %d. character position\n%s\n", u8 - src, src);    
            u2->h = 0xff;
            u2->l = 0xfd;
            break;
        }
        case 0xc0:
        case 0xd0: {    // 2-byte UTF-8 codes
            if ((*(u8+1) & 0xc0) == 0x80) {
	        u2->h = (*u8 & 0x1f) >> 2;
                u2->l = (*u8 << 6) + (*(u8+1) & 0x3f);
	        u8++;
            } else {
                fprintf(stderr, "UTF-8 encoding error. Missing continuation byte in %d. character position:\n%s\n", u8 - src, src);
                u2->h = 0xff;
                u2->l = 0xfd;
            }
            break;
        }
        case 0xe0: {    // 3-byte UTF-8 codes
            if ((*(u8+1) & 0xc0) == 0x80) {
	        u2->h = ((*u8 & 0x0f) << 4) + ((*(u8+1) & 0x3f) >> 2);
                u8++;
                if ((*(u8+1) & 0xc0) == 0x80) {
                    u2->l = (*u8 << 6) + (*(u8+1) & 0x3f);
	            u8++;
                } else {
                    fprintf(stderr, "UTF-8 encoding error. Missing continuation byte in %d. character position:\n%s\n", u8 - src, src);
                    u2->h = 0xff;
                    u2->l = 0xfd;
                }
            } else {
                fprintf(stderr, "UTF-8 encoding error. Missing continuation byte in %d. character position:\n%s\n", u8 - src, src);
                u2->h = 0xff;
                u2->l = 0xfd;
            }
            break;
        }
        case 0xf0: {    // 4 or more byte UTF-8 codes
            fprintf(stderr, "This UTF-8 encoding can't convert to UTF-16:\n%s\n", src);
            u2->h = 0xff;
            u2->l = 0xfd;
            break;            
        }
    }
    u8++;
    u2++;
    }
    return u2 - dest;
}

void flag_qsort(unsigned short flags[], int begin, int end) {
    unsigned short reg;
    if (end > begin) {
        unsigned short pivot = flags[begin];
        int l = begin + 1;
        int r = end;
        while(l < r) {
            if (flags[l] <= pivot) {
                l++;
            } else {
             r--;
	     reg = flags[l];
	     flags[l] = flags[r];
	     flags[r] = reg;
          }
       }
       l--;
       reg = flags[begin];
       flags[begin] = flags[l];
       flags[l] = reg;

       flag_qsort(flags, begin, l);
       flag_qsort(flags, r, end);
    }
 }

int flag_bsearch(unsigned short flags[], unsigned short flag, int length) {
   int mid;
   int left = 0;
   int right = length - 1;
   while (left <= right) {
      mid = (left + right) / 2;
      if (flags[mid] == flag) return 1;
      if (flag < flags[mid]) right = mid - 1;
      else left = mid + 1;
   }
   return 0;
}

 // strip strings into token based on single char delimiter
 // acts like strsep() but only uses a delim char and not
 // a delim string
 // default delimiter: white space characters
 
 char * mystrsep(char ** stringp, const char delim)
 {
   char * rv = NULL;
   char * mp = *stringp;
   int n = strlen(mp);
   if (n > 0) {
      char * dp;
      if (delim) {
        dp = (char *)memchr(mp,(int)((unsigned char)delim),n);
      } else {
	// don't use isspace() here, the string can be in some random charset
	// that's way different than the locale's
        for (dp = mp; (*dp && *dp != ' ' && *dp != '\t'); dp++);
        if (!*dp) dp = NULL;
      }
      if (dp) {
         *stringp = dp+1;
         int nc = (int)((unsigned long)dp - (unsigned long)mp);
         rv = (char *) malloc(nc+1);
         memcpy(rv,mp,nc);
         *(rv+nc) = '\0';
         return rv;
      } else {
        rv = (char *) malloc(n+1);
        memcpy(rv, mp, n);
        *(rv+n) = '\0';
        *stringp = mp + n;
        return rv;
      }
   }
   return NULL;
 }

 
 // replaces strdup with ansi version
 char * mystrdup(const char * s)
 {
   char * d = NULL;
   if (s) {
      int sl = strlen(s);
      d = (char *) malloc(((sl+1) * sizeof(char)));
      if (d) memcpy(d,s,((sl+1)*sizeof(char)));
   }
   return d;
 }
 
 
 // remove cross-platform text line end characters
 void mychomp(char * s)
 {
   int k = strlen(s);
   if ((k > 0) && ((*(s+k-1)=='\r') || (*(s+k-1)=='\n'))) *(s+k-1) = '\0';
   if ((k > 1) && (*(s+k-2) == '\r')) *(s+k-2) = '\0';
 }
 
 
 //  does an ansi strdup of the reverse of a string
 char * myrevstrdup(const char * s)
 {
     char * d = NULL;
     if (s) {
        int sl = strlen(s);
        d = (char *) malloc((sl+1) * sizeof(char));
        if (d) {
          const char * p = s + sl - 1;
          char * q = d;
          while (p >= s) *q++ = *p--;
          *q = '\0';
        }
     }
     return d;
 }

 // convert null terminated string to all caps using encoding
 void enmkallcap(char * d, const char * p, const char * encoding)
 
 {
   struct cs_info * csconv = get_current_cs(encoding);
   while (*p != '\0') {
     *d++ = csconv[((unsigned char) *p)].cupper;
     p++;
   }
   *d = '\0';
 }


 // append s to ends of every lines in text
 void strlinecat(char * dest, const char * s)
 {
    char * dup = mystrdup(dest);
    char * source = dup;
    int len = strlen(s);
    while (*source) {
        if (*source == '\n') {
            strncpy(dest, s, len);
            dest += len;
        }
        *dest = *source;
        source++; dest++;
    }
    strcpy(dest, s);
    free(dup);
 }

// break text to lines
// return number of lines
int line_tok(const char * text, char *** lines) {
    int linenum = 0;
    char * dup = mystrdup(text);
    char * p = dup;
    while ((p = strchr(p, '\n'))) {
        linenum++;
        *p = '\0';
        p++;
        if (*p == '\0') break;
    }
    *lines = (char **) calloc(linenum + 1, sizeof(char *));
    if (!(*lines)) return -1;

    p = dup; 
    for (int i = 0; i < linenum + 1; i++) {
        (*lines)[i] = mystrdup(p);
        p += strlen(p) + 1;
    }
    free(dup);
    return linenum;
}

// uniq line in place
char * line_uniq(char * text) {
    char ** lines;
    char linenum = line_tok(text, &lines);
    int i;
    strcpy(text, lines[0]);
    for ( i = 1; i<=linenum; i++ ) {
        int dup = 0;
        for (int j = 0; j < i; j++) {
            if (strcmp(lines[i], lines[j]) == 0) dup = 1;
        }
        if (!dup) {
            if ((i > 1) || (*(lines[0]) != '\0')) strcat(text, "\n");
            strcat(text, lines[i]);
        }
    }
    for ( i = 0; i<=linenum; i++ ) {
        if (lines[i]) free(lines[i]);
    }
    if (lines) free(lines);
    return text;
}

// change \n to char c
char * line_join(char * text, char c) {
    char * p;
    for (p = text; *p; p++) if (*p == '\n') *p = c;
    return text;
}

// leave only last {[^}]*} substring for handling zero morphemes
char * delete_zeros(char * morphout) {
    char * p = morphout;
    char * q = p;
    char * q2 = NULL;
    int suffix = 0;
    
    for (;*p && *(p+1);) {
        switch (*p) {
            case '{': 
                q2 = q;
                q--;
                break;
            case '}':
                if (q2) {
                    suffix = 1;
                    q--;
                }
                break; 
            default:
                if (suffix) {
                    q = q2;
                }
                suffix = 0;
                *q = *p;
        }
        p++;
        q++;
    }
    *q = '\0';
    return morphout;
}

char * mystrrep(char * word, const char * pat, const char * rep) {
    char * pos = strstr(word, pat);
    if (pos) {
        int replen = strlen(rep);
        int patlen = strlen(pat);
        if (replen < patlen) {
            char * end = word + strlen(word);
            char * next = pos + replen;
            char * prev = pos + strlen(pat);
            for (; prev < end; *next = *prev, prev++, next++);
            *next = '\0';
        } else if (replen > patlen) {
            char * end = pos + patlen;
            char * next = word + strlen(word) + replen - patlen;
            char * prev = next - replen + patlen;
            for (; prev >= end; *next = *prev, prev--, next--);
        }
        strncpy(pos, rep, replen);
    }
    return word;
}

 // convert null terminated string to all little using encoding
 void enmkallsmall(char * d, const char * p, const char * encoding)
 {
   struct cs_info * csconv = get_current_cs(encoding);
   while (*p != '\0') {
     *d++ = csconv[((unsigned char) *p)].clower;
     p++;
   }
   *d = '\0';
 }
 // convert null terminated string to have intial capital using encoding
 void enmkinitcap(char * d, const char * p, const char * encoding)
 {
   struct cs_info * csconv = get_current_cs(encoding);
   memcpy(d,p,(strlen(p)+1));
   if (*p != '\0') *d= csconv[((unsigned char)*p)].cupper;
 }
 
 
 // convert null terminated string to all caps
 void mkallcap(char * p, const struct cs_info * csconv)
 {
   while (*p != '\0') {
     *p = csconv[((unsigned char) *p)].cupper;
     p++;
   }
 }
 
 
 // convert null terminated string to all little
 void mkallsmall(char * p, const struct cs_info * csconv)
 {
   while (*p != '\0') {
     *p = csconv[((unsigned char) *p)].clower;
     p++;
   }
 }

void mkallsmall_utf(w_char * u, int nc, struct unicode_info2 * utfconv) {
    for (int i = 0; i < nc; i++) {
        unsigned short idx = (u[i].h << 8) + u[i].l;
        if (idx != utfconv[idx].clower) {
            u[i].h = (unsigned char) (utfconv[idx].clower >> 8);
            u[i].l = (unsigned char) (utfconv[idx].clower & 0x00FF);
        }
    }
}
 
 // convert null terminated string to have intial capital
 void mkinitcap(char * p, const struct cs_info * csconv)
 {
   if (*p != '\0') *p = csconv[((unsigned char)*p)].cupper;
 }

 // reverse word 
 void reverseword(char * word) {
   char r;
   for (char * dest = word + strlen(word) - 1; word < dest; word++, dest--) {
     r=*word;
     *word = *dest;
     *dest = r;
   }
 }

 // reverse word 
 void reverseword_utf(char * word) {
   w_char w[MAXWORDLEN];
   w_char * p;
   w_char r;
   int l = u8_u16(w, MAXWORDLEN, word);
   p = w;
   for (w_char * dest = w + l - 1; p < dest; p++, dest--) {
     r=*p;
     *p = *dest;
     *dest = r;
   }
   u16_u8(word, MAXWORDUTF8LEN, w, l);
 }

// these are simple character mappings for the 
// encodings supported
// supplying isupper, tolower, and toupper

struct cs_info iso1_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xff },
};


struct cs_info iso2_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x01, 0xb1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x01, 0xb3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x01, 0xb5, 0xa5 },
{ 0x01, 0xb6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x01, 0xb9, 0xa9 },
{ 0x01, 0xba, 0xaa },
{ 0x01, 0xbb, 0xab },
{ 0x01, 0xbc, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x01, 0xbe, 0xae },
{ 0x01, 0xbf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xa1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xa3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xa5 },
{ 0x00, 0xb6, 0xa6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xa9 },
{ 0x00, 0xba, 0xaa },
{ 0x00, 0xbb, 0xab },
{ 0x00, 0xbc, 0xac },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xae },
{ 0x00, 0xbf, 0xaf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xff },
};


struct cs_info iso3_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x01, 0xb1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x01, 0xb6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x01, 0x69, 0xa9 },
{ 0x01, 0xba, 0xaa },
{ 0x01, 0xbb, 0xab },
{ 0x01, 0xbc, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x01, 0xbf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xa1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xa6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0x49 },
{ 0x00, 0xba, 0xaa },
{ 0x00, 0xbb, 0xab },
{ 0x00, 0xbc, 0xac },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xaf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x00, 0xc3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x00, 0xd0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xe3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xf0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso4_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x01, 0xb1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x01, 0xb3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x01, 0xb5, 0xa5 },
{ 0x01, 0xb6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x01, 0xb9, 0xa9 },
{ 0x01, 0xba, 0xaa },
{ 0x01, 0xbb, 0xab },
{ 0x01, 0xbc, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x01, 0xbe, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xa1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xa3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xa5 },
{ 0x00, 0xb6, 0xa6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xa9 },
{ 0x00, 0xba, 0xaa },
{ 0x00, 0xbb, 0xab },
{ 0x00, 0xbc, 0xac },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xae },
{ 0x00, 0xbf, 0xbf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso5_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x01, 0xf1, 0xa1 },
{ 0x01, 0xf2, 0xa2 },
{ 0x01, 0xf3, 0xa3 },
{ 0x01, 0xf4, 0xa4 },
{ 0x01, 0xf5, 0xa5 },
{ 0x01, 0xf6, 0xa6 },
{ 0x01, 0xf7, 0xa7 },
{ 0x01, 0xf8, 0xa8 },
{ 0x01, 0xf9, 0xa9 },
{ 0x01, 0xfa, 0xaa },
{ 0x01, 0xfb, 0xab },
{ 0x01, 0xfc, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x01, 0xfe, 0xae },
{ 0x01, 0xff, 0xaf },
{ 0x01, 0xd0, 0xb0 },
{ 0x01, 0xd1, 0xb1 },
{ 0x01, 0xd2, 0xb2 },
{ 0x01, 0xd3, 0xb3 },
{ 0x01, 0xd4, 0xb4 },
{ 0x01, 0xd5, 0xb5 },
{ 0x01, 0xd6, 0xb6 },
{ 0x01, 0xd7, 0xb7 },
{ 0x01, 0xd8, 0xb8 },
{ 0x01, 0xd9, 0xb9 },
{ 0x01, 0xda, 0xba },
{ 0x01, 0xdb, 0xbb },
{ 0x01, 0xdc, 0xbc },
{ 0x01, 0xdd, 0xbd },
{ 0x01, 0xde, 0xbe },
{ 0x01, 0xdf, 0xbf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x00, 0xd0, 0xb0 },
{ 0x00, 0xd1, 0xb1 },
{ 0x00, 0xd2, 0xb2 },
{ 0x00, 0xd3, 0xb3 },
{ 0x00, 0xd4, 0xb4 },
{ 0x00, 0xd5, 0xb5 },
{ 0x00, 0xd6, 0xb6 },
{ 0x00, 0xd7, 0xb7 },
{ 0x00, 0xd8, 0xb8 },
{ 0x00, 0xd9, 0xb9 },
{ 0x00, 0xda, 0xba },
{ 0x00, 0xdb, 0xbb },
{ 0x00, 0xdc, 0xbc },
{ 0x00, 0xdd, 0xbd },
{ 0x00, 0xde, 0xbe },
{ 0x00, 0xdf, 0xbf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xf0 },
{ 0x00, 0xf1, 0xa1 },
{ 0x00, 0xf2, 0xa2 },
{ 0x00, 0xf3, 0xa3 },
{ 0x00, 0xf4, 0xa4 },
{ 0x00, 0xf5, 0xa5 },
{ 0x00, 0xf6, 0xa6 },
{ 0x00, 0xf7, 0xa7 },
{ 0x00, 0xf8, 0xa8 },
{ 0x00, 0xf9, 0xa9 },
{ 0x00, 0xfa, 0xaa },
{ 0x00, 0xfb, 0xab },
{ 0x00, 0xfc, 0xac },
{ 0x00, 0xfd, 0xfd },
{ 0x00, 0xfe, 0xae },
{ 0x00, 0xff, 0xaf },
};

struct cs_info iso6_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x00, 0xc0, 0xc0 },
{ 0x00, 0xc1, 0xc1 },
{ 0x00, 0xc2, 0xc2 },
{ 0x00, 0xc3, 0xc3 },
{ 0x00, 0xc4, 0xc4 },
{ 0x00, 0xc5, 0xc5 },
{ 0x00, 0xc6, 0xc6 },
{ 0x00, 0xc7, 0xc7 },
{ 0x00, 0xc8, 0xc8 },
{ 0x00, 0xc9, 0xc9 },
{ 0x00, 0xca, 0xca },
{ 0x00, 0xcb, 0xcb },
{ 0x00, 0xcc, 0xcc },
{ 0x00, 0xcd, 0xcd },
{ 0x00, 0xce, 0xce },
{ 0x00, 0xcf, 0xcf },
{ 0x00, 0xd0, 0xd0 },
{ 0x00, 0xd1, 0xd1 },
{ 0x00, 0xd2, 0xd2 },
{ 0x00, 0xd3, 0xd3 },
{ 0x00, 0xd4, 0xd4 },
{ 0x00, 0xd5, 0xd5 },
{ 0x00, 0xd6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x00, 0xd8, 0xd8 },
{ 0x00, 0xd9, 0xd9 },
{ 0x00, 0xda, 0xda },
{ 0x00, 0xdb, 0xdb },
{ 0x00, 0xdc, 0xdc },
{ 0x00, 0xdd, 0xdd },
{ 0x00, 0xde, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xe0 },
{ 0x00, 0xe1, 0xe1 },
{ 0x00, 0xe2, 0xe2 },
{ 0x00, 0xe3, 0xe3 },
{ 0x00, 0xe4, 0xe4 },
{ 0x00, 0xe5, 0xe5 },
{ 0x00, 0xe6, 0xe6 },
{ 0x00, 0xe7, 0xe7 },
{ 0x00, 0xe8, 0xe8 },
{ 0x00, 0xe9, 0xe9 },
{ 0x00, 0xea, 0xea },
{ 0x00, 0xeb, 0xeb },
{ 0x00, 0xec, 0xec },
{ 0x00, 0xed, 0xed },
{ 0x00, 0xee, 0xee },
{ 0x00, 0xef, 0xef },
{ 0x00, 0xf0, 0xf0 },
{ 0x00, 0xf1, 0xf1 },
{ 0x00, 0xf2, 0xf2 },
{ 0x00, 0xf3, 0xf3 },
{ 0x00, 0xf4, 0xf4 },
{ 0x00, 0xf5, 0xf5 },
{ 0x00, 0xf6, 0xf6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xf8 },
{ 0x00, 0xf9, 0xf9 },
{ 0x00, 0xfa, 0xfa },
{ 0x00, 0xfb, 0xfb },
{ 0x00, 0xfc, 0xfc },
{ 0x00, 0xfd, 0xfd },
{ 0x00, 0xfe, 0xfe },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso7_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x01, 0xdc, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x01, 0xdd, 0xb8 },
{ 0x01, 0xde, 0xb9 },
{ 0x01, 0xdf, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x01, 0xfc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x01, 0xfd, 0xbe },
{ 0x01, 0xfe, 0xbf },
{ 0x00, 0xc0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x00, 0xd2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x01, 0xf7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x00, 0xdc, 0xb6 },
{ 0x00, 0xdd, 0xb8 },
{ 0x00, 0xde, 0xb9 },
{ 0x00, 0xdf, 0xba },
{ 0x00, 0xe0, 0xe0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd3 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xd7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xbc },
{ 0x00, 0xfd, 0xbe },
{ 0x00, 0xfe, 0xbf },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso8_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x00, 0xc0, 0xc0 },
{ 0x00, 0xc1, 0xc1 },
{ 0x00, 0xc2, 0xc2 },
{ 0x00, 0xc3, 0xc3 },
{ 0x00, 0xc4, 0xc4 },
{ 0x00, 0xc5, 0xc5 },
{ 0x00, 0xc6, 0xc6 },
{ 0x00, 0xc7, 0xc7 },
{ 0x00, 0xc8, 0xc8 },
{ 0x00, 0xc9, 0xc9 },
{ 0x00, 0xca, 0xca },
{ 0x00, 0xcb, 0xcb },
{ 0x00, 0xcc, 0xcc },
{ 0x00, 0xcd, 0xcd },
{ 0x00, 0xce, 0xce },
{ 0x00, 0xcf, 0xcf },
{ 0x00, 0xd0, 0xd0 },
{ 0x00, 0xd1, 0xd1 },
{ 0x00, 0xd2, 0xd2 },
{ 0x00, 0xd3, 0xd3 },
{ 0x00, 0xd4, 0xd4 },
{ 0x00, 0xd5, 0xd5 },
{ 0x00, 0xd6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x00, 0xd8, 0xd8 },
{ 0x00, 0xd9, 0xd9 },
{ 0x00, 0xda, 0xda },
{ 0x00, 0xdb, 0xdb },
{ 0x00, 0xdc, 0xdc },
{ 0x00, 0xdd, 0xdd },
{ 0x00, 0xde, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xe0 },
{ 0x00, 0xe1, 0xe1 },
{ 0x00, 0xe2, 0xe2 },
{ 0x00, 0xe3, 0xe3 },
{ 0x00, 0xe4, 0xe4 },
{ 0x00, 0xe5, 0xe5 },
{ 0x00, 0xe6, 0xe6 },
{ 0x00, 0xe7, 0xe7 },
{ 0x00, 0xe8, 0xe8 },
{ 0x00, 0xe9, 0xe9 },
{ 0x00, 0xea, 0xea },
{ 0x00, 0xeb, 0xeb },
{ 0x00, 0xec, 0xec },
{ 0x00, 0xed, 0xed },
{ 0x00, 0xee, 0xee },
{ 0x00, 0xef, 0xef },
{ 0x00, 0xf0, 0xf0 },
{ 0x00, 0xf1, 0xf1 },
{ 0x00, 0xf2, 0xf2 },
{ 0x00, 0xf3, 0xf3 },
{ 0x00, 0xf4, 0xf4 },
{ 0x00, 0xf5, 0xf5 },
{ 0x00, 0xf6, 0xf6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xf8 },
{ 0x00, 0xf9, 0xf9 },
{ 0x00, 0xfa, 0xfa },
{ 0x00, 0xfb, 0xfb },
{ 0x00, 0xfc, 0xfc },
{ 0x00, 0xfd, 0xfd },
{ 0x00, 0xfe, 0xfe },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso9_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0xfd, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0xdd },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0x69, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0x49 },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso10_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x00, 0xc0, 0xc0 },
{ 0x00, 0xc1, 0xc1 },
{ 0x00, 0xc2, 0xc2 },
{ 0x00, 0xc3, 0xc3 },
{ 0x00, 0xc4, 0xc4 },
{ 0x00, 0xc5, 0xc5 },
{ 0x00, 0xc6, 0xc6 },
{ 0x00, 0xc7, 0xc7 },
{ 0x00, 0xc8, 0xc8 },
{ 0x00, 0xc9, 0xc9 },
{ 0x00, 0xca, 0xca },
{ 0x00, 0xcb, 0xcb },
{ 0x00, 0xcc, 0xcc },
{ 0x00, 0xcd, 0xcd },
{ 0x00, 0xce, 0xce },
{ 0x00, 0xcf, 0xcf },
{ 0x00, 0xd0, 0xd0 },
{ 0x00, 0xd1, 0xd1 },
{ 0x00, 0xd2, 0xd2 },
{ 0x00, 0xd3, 0xd3 },
{ 0x00, 0xd4, 0xd4 },
{ 0x00, 0xd5, 0xd5 },
{ 0x00, 0xd6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x00, 0xd8, 0xd8 },
{ 0x00, 0xd9, 0xd9 },
{ 0x00, 0xda, 0xda },
{ 0x00, 0xdb, 0xdb },
{ 0x00, 0xdc, 0xdc },
{ 0x00, 0xdd, 0xdd },
{ 0x00, 0xde, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xe0 },
{ 0x00, 0xe1, 0xe1 },
{ 0x00, 0xe2, 0xe2 },
{ 0x00, 0xe3, 0xe3 },
{ 0x00, 0xe4, 0xe4 },
{ 0x00, 0xe5, 0xe5 },
{ 0x00, 0xe6, 0xe6 },
{ 0x00, 0xe7, 0xe7 },
{ 0x00, 0xe8, 0xe8 },
{ 0x00, 0xe9, 0xe9 },
{ 0x00, 0xea, 0xea },
{ 0x00, 0xeb, 0xeb },
{ 0x00, 0xec, 0xec },
{ 0x00, 0xed, 0xed },
{ 0x00, 0xee, 0xee },
{ 0x00, 0xef, 0xef },
{ 0x00, 0xf0, 0xf0 },
{ 0x00, 0xf1, 0xf1 },
{ 0x00, 0xf2, 0xf2 },
{ 0x00, 0xf3, 0xf3 },
{ 0x00, 0xf4, 0xf4 },
{ 0x00, 0xf5, 0xf5 },
{ 0x00, 0xf6, 0xf6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xf8 },
{ 0x00, 0xf9, 0xf9 },
{ 0x00, 0xfa, 0xfa },
{ 0x00, 0xfb, 0xfb },
{ 0x00, 0xfc, 0xfc },
{ 0x00, 0xfd, 0xfd },
{ 0x00, 0xfe, 0xfe },
{ 0x00, 0xff, 0xff },
};

struct cs_info koi8r_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xb3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x01, 0xa3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x00, 0xc0, 0xe0 },
{ 0x00, 0xc1, 0xe1 },
{ 0x00, 0xc2, 0xe2 },
{ 0x00, 0xc3, 0xe3 },
{ 0x00, 0xc4, 0xe4 },
{ 0x00, 0xc5, 0xe5 },
{ 0x00, 0xc6, 0xe6 },
{ 0x00, 0xc7, 0xe7 },
{ 0x00, 0xc8, 0xe8 },
{ 0x00, 0xc9, 0xe9 },
{ 0x00, 0xca, 0xea },
{ 0x00, 0xcb, 0xeb },
{ 0x00, 0xcc, 0xec },
{ 0x00, 0xcd, 0xed },
{ 0x00, 0xce, 0xee },
{ 0x00, 0xcf, 0xef },
{ 0x00, 0xd0, 0xf0 },
{ 0x00, 0xd1, 0xf1 },
{ 0x00, 0xd2, 0xf2 },
{ 0x00, 0xd3, 0xf3 },
{ 0x00, 0xd4, 0xf4 },
{ 0x00, 0xd5, 0xf5 },
{ 0x00, 0xd6, 0xf6 },
{ 0x00, 0xd7, 0xf7 },
{ 0x00, 0xd8, 0xf8 },
{ 0x00, 0xd9, 0xf9 },
{ 0x00, 0xda, 0xfa },
{ 0x00, 0xdb, 0xfb },
{ 0x00, 0xdc, 0xfc },
{ 0x00, 0xdd, 0xfd },
{ 0x00, 0xde, 0xfe },
{ 0x00, 0xdf, 0xff },
{ 0x01, 0xc0, 0xe0 },
{ 0x01, 0xc1, 0xe1 },
{ 0x01, 0xc2, 0xe2 },
{ 0x01, 0xc3, 0xe3 },
{ 0x01, 0xc4, 0xe4 },
{ 0x01, 0xc5, 0xe5 },
{ 0x01, 0xc6, 0xe6 },
{ 0x01, 0xc7, 0xe7 },
{ 0x01, 0xc8, 0xe8 },
{ 0x01, 0xc9, 0xe9 },
{ 0x01, 0xca, 0xea },
{ 0x01, 0xcb, 0xeb },
{ 0x01, 0xcc, 0xec },
{ 0x01, 0xcd, 0xed },
{ 0x01, 0xce, 0xee },
{ 0x01, 0xcf, 0xef },
{ 0x01, 0xd0, 0xf0 },
{ 0x01, 0xd1, 0xf1 },
{ 0x01, 0xd2, 0xf2 },
{ 0x01, 0xd3, 0xf3 },
{ 0x01, 0xd4, 0xf4 },
{ 0x01, 0xd5, 0xf5 },
{ 0x01, 0xd6, 0xf6 },
{ 0x01, 0xd7, 0xf7 },
{ 0x01, 0xd8, 0xf8 },
{ 0x01, 0xd9, 0xf9 },
{ 0x01, 0xda, 0xfa },
{ 0x01, 0xdb, 0xfb },
{ 0x01, 0xdc, 0xfc },
{ 0x01, 0xdd, 0xfd },
{ 0x01, 0xde, 0xfe },
{ 0x01, 0xdf, 0xff },
};

struct cs_info koi8u_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xb3 },
{ 0x00, 0xa4, 0xb4 }, /* ie */
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xb6 }, /* i */
{ 0x00, 0xa7, 0xb7 }, /* ii */
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xbd }, /* g'' */
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x01, 0xa3, 0xb3 },
{ 0x00, 0xb4, 0xb4 }, /* IE */
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 }, /* I */
{ 0x00, 0xb7, 0xb7 }, /* II */
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x00, 0xc0, 0xe0 },
{ 0x00, 0xc1, 0xe1 },
{ 0x00, 0xc2, 0xe2 },
{ 0x00, 0xc3, 0xe3 },
{ 0x00, 0xc4, 0xe4 },
{ 0x00, 0xc5, 0xe5 },
{ 0x00, 0xc6, 0xe6 },
{ 0x00, 0xc7, 0xe7 },
{ 0x00, 0xc8, 0xe8 },
{ 0x00, 0xc9, 0xe9 },
{ 0x00, 0xca, 0xea },
{ 0x00, 0xcb, 0xeb },
{ 0x00, 0xcc, 0xec },
{ 0x00, 0xcd, 0xed },
{ 0x00, 0xce, 0xee },
{ 0x00, 0xcf, 0xef },
{ 0x00, 0xd0, 0xf0 },
{ 0x00, 0xd1, 0xf1 },
{ 0x00, 0xd2, 0xf2 },
{ 0x00, 0xd3, 0xf3 },
{ 0x00, 0xd4, 0xf4 },
{ 0x00, 0xd5, 0xf5 },
{ 0x00, 0xd6, 0xf6 },
{ 0x00, 0xd7, 0xf7 },
{ 0x00, 0xd8, 0xf8 },
{ 0x00, 0xd9, 0xf9 },
{ 0x00, 0xda, 0xfa },
{ 0x00, 0xdb, 0xfb },
{ 0x00, 0xdc, 0xfc },
{ 0x00, 0xdd, 0xfd },
{ 0x00, 0xde, 0xfe },
{ 0x00, 0xdf, 0xff },
{ 0x01, 0xc0, 0xe0 },
{ 0x01, 0xc1, 0xe1 },
{ 0x01, 0xc2, 0xe2 },
{ 0x01, 0xc3, 0xe3 },
{ 0x01, 0xc4, 0xe4 },
{ 0x01, 0xc5, 0xe5 },
{ 0x01, 0xc6, 0xe6 },
{ 0x01, 0xc7, 0xe7 },
{ 0x01, 0xc8, 0xe8 },
{ 0x01, 0xc9, 0xe9 },
{ 0x01, 0xca, 0xea },
{ 0x01, 0xcb, 0xeb },
{ 0x01, 0xcc, 0xec },
{ 0x01, 0xcd, 0xed },
{ 0x01, 0xce, 0xee },
{ 0x01, 0xcf, 0xef },
{ 0x01, 0xd0, 0xf0 },
{ 0x01, 0xd1, 0xf1 },
{ 0x01, 0xd2, 0xf2 },
{ 0x01, 0xd3, 0xf3 },
{ 0x01, 0xd4, 0xf4 },
{ 0x01, 0xd5, 0xf5 },
{ 0x01, 0xd6, 0xf6 },
{ 0x01, 0xd7, 0xf7 },
{ 0x01, 0xd8, 0xf8 },
{ 0x01, 0xd9, 0xf9 },
{ 0x01, 0xda, 0xfa },
{ 0x01, 0xdb, 0xfb },
{ 0x01, 0xdc, 0xfc },
{ 0x01, 0xdd, 0xfd },
{ 0x01, 0xde, 0xfe },
{ 0x01, 0xdf, 0xff },
};

struct cs_info cp1251_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x01, 0x90, 0x80 },
{ 0x01, 0x83, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x81 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x01, 0x9a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x01, 0x9c, 0x8c },
{ 0x01, 0x9d, 0x8d },
{ 0x01, 0x9e, 0x8e },
{ 0x01, 0x9f, 0x8f },
{ 0x00, 0x90, 0x80 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x8a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x8c },
{ 0x00, 0x9d, 0x8d },
{ 0x00, 0x9e, 0x8e },
{ 0x00, 0x9f, 0x8f },
{ 0x00, 0xa0, 0xa0 },
{ 0x01, 0xa2, 0xa1 },
{ 0x00, 0xa2, 0xa1 },
{ 0x01, 0xbc, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x01, 0xb4, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x01, 0xb8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x01, 0xba, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x01, 0xbf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x01, 0xb3, 0xb2 },
{ 0x00, 0xb3, 0xb2 },
{ 0x00, 0xb4, 0xa5 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xa8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xaa },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xa3 },
{ 0x01, 0xbe, 0xbd },
{ 0x00, 0xbe, 0xbd },
{ 0x00, 0xbf, 0xaf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x01, 0xf7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x01, 0xff, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xd7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xdf },
};

struct cs_info iso13_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0A, 0x0A },
{ 0x00, 0x0B, 0x0B },
{ 0x00, 0x0C, 0x0C },
{ 0x00, 0x0D, 0x0D },
{ 0x00, 0x0E, 0x0E },
{ 0x00, 0x0F, 0x0F },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1A, 0x1A },
{ 0x00, 0x1B, 0x1B },
{ 0x00, 0x1C, 0x1C },
{ 0x00, 0x1D, 0x1D },
{ 0x00, 0x1E, 0x1E },
{ 0x00, 0x1F, 0x1F },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2A, 0x2A },
{ 0x00, 0x2B, 0x2B },
{ 0x00, 0x2C, 0x2C },
{ 0x00, 0x2D, 0x2D },
{ 0x00, 0x2E, 0x2E },
{ 0x00, 0x2F, 0x2F },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3A, 0x3A },
{ 0x00, 0x3B, 0x3B },
{ 0x00, 0x3C, 0x3C },
{ 0x00, 0x3D, 0x3D },
{ 0x00, 0x3E, 0x3E },
{ 0x00, 0x3F, 0x3F },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6A, 0x4A },
{ 0x01, 0x6B, 0x4B },
{ 0x01, 0x6C, 0x4C },
{ 0x01, 0x6D, 0x4D },
{ 0x01, 0x6E, 0x4E },
{ 0x01, 0x6F, 0x4F },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7A, 0x5A },
{ 0x00, 0x5B, 0x5B },
{ 0x00, 0x5C, 0x5C },
{ 0x00, 0x5D, 0x5D },
{ 0x00, 0x5E, 0x5E },
{ 0x00, 0x5F, 0x5F },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6A, 0x4A },
{ 0x00, 0x6B, 0x4B },
{ 0x00, 0x6C, 0x4C },
{ 0x00, 0x6D, 0x4D },
{ 0x00, 0x6E, 0x4E },
{ 0x00, 0x6F, 0x4F },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7A, 0x5A },
{ 0x00, 0x7B, 0x7B },
{ 0x00, 0x7C, 0x7C },
{ 0x00, 0x7D, 0x7D },
{ 0x00, 0x7E, 0x7E },
{ 0x00, 0x7F, 0x7F },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8A, 0x8A },
{ 0x00, 0x8B, 0x8B },
{ 0x00, 0x8C, 0x8C },
{ 0x00, 0x8D, 0x8D },
{ 0x00, 0x8E, 0x8E },
{ 0x00, 0x8F, 0x8F },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9A, 0x9A },
{ 0x00, 0x9B, 0x9B },
{ 0x00, 0x9C, 0x9C },
{ 0x00, 0x9D, 0x9D },
{ 0x00, 0x9E, 0x9E },
{ 0x00, 0x9F, 0x9F },
{ 0x00, 0xA0, 0xA0 },
{ 0x00, 0xA1, 0xA1 },
{ 0x00, 0xA2, 0xA2 },
{ 0x00, 0xA3, 0xA3 },
{ 0x00, 0xA4, 0xA4 },
{ 0x00, 0xA5, 0xA5 },
{ 0x00, 0xA6, 0xA6 },
{ 0x00, 0xA7, 0xA7 },
{ 0x01, 0xB8, 0xA8 },
{ 0x00, 0xA9, 0xA9 },
{ 0x01, 0xBA, 0xAA },
{ 0x00, 0xAB, 0xAB },
{ 0x00, 0xAC, 0xAC },
{ 0x00, 0xAD, 0xAD },
{ 0x00, 0xAE, 0xAE },
{ 0x01, 0xBF, 0xAF },
{ 0x00, 0xB0, 0xB0 },
{ 0x00, 0xB1, 0xB1 },
{ 0x00, 0xB2, 0xB2 },
{ 0x00, 0xB3, 0xB3 },
{ 0x00, 0xB4, 0xB4 },
{ 0x00, 0xB5, 0xB5 },
{ 0x00, 0xB6, 0xB6 },
{ 0x00, 0xB7, 0xB7 },
{ 0x00, 0xB8, 0xA8 },
{ 0x00, 0xB9, 0xB9 },
{ 0x00, 0xBA, 0xAA },
{ 0x00, 0xBB, 0xBB },
{ 0x00, 0xBC, 0xBC },
{ 0x00, 0xBD, 0xBD },
{ 0x00, 0xBE, 0xBE },
{ 0x00, 0xBF, 0xAF },
{ 0x01, 0xE0, 0xC0 },
{ 0x01, 0xE1, 0xC1 },
{ 0x01, 0xE2, 0xC2 },
{ 0x01, 0xE3, 0xC3 },
{ 0x01, 0xE4, 0xC4 },
{ 0x01, 0xE5, 0xC5 },
{ 0x01, 0xE6, 0xC6 },
{ 0x01, 0xE7, 0xC7 },
{ 0x01, 0xE8, 0xC8 },
{ 0x01, 0xE9, 0xC9 },
{ 0x01, 0xEA, 0xCA },
{ 0x01, 0xEB, 0xCB },
{ 0x01, 0xEC, 0xCC },
{ 0x01, 0xED, 0xCD },
{ 0x01, 0xEE, 0xCE },
{ 0x01, 0xEF, 0xCF },
{ 0x01, 0xF0, 0xD0 },
{ 0x01, 0xF1, 0xD1 },
{ 0x01, 0xF2, 0xD2 },
{ 0x01, 0xF3, 0xD3 },
{ 0x01, 0xF4, 0xD4 },
{ 0x01, 0xF5, 0xD5 },
{ 0x01, 0xF6, 0xD6 },
{ 0x00, 0xD7, 0xD7 },
{ 0x01, 0xF8, 0xD8 },
{ 0x01, 0xF9, 0xD9 },
{ 0x01, 0xFA, 0xDA },
{ 0x01, 0xFB, 0xDB },
{ 0x01, 0xFC, 0xDC },
{ 0x01, 0xFD, 0xDD },
{ 0x01, 0xFE, 0xDE },
{ 0x00, 0xDF, 0xDF },
{ 0x00, 0xE0, 0xC0 },
{ 0x00, 0xE1, 0xC1 },
{ 0x00, 0xE2, 0xC2 },
{ 0x00, 0xE3, 0xC3 },
{ 0x00, 0xE4, 0xC4 },
{ 0x00, 0xE5, 0xC5 },
{ 0x00, 0xE6, 0xC6 },
{ 0x00, 0xE7, 0xC7 },
{ 0x00, 0xE8, 0xC8 },
{ 0x00, 0xE9, 0xC9 },
{ 0x00, 0xEA, 0xCA },
{ 0x00, 0xEB, 0xCB },
{ 0x00, 0xEC, 0xCC },
{ 0x00, 0xED, 0xCD },
{ 0x00, 0xEE, 0xCE },
{ 0x00, 0xEF, 0xCF },
{ 0x00, 0xF0, 0xD0 },
{ 0x00, 0xF1, 0xD1 },
{ 0x00, 0xF2, 0xD2 },
{ 0x00, 0xF3, 0xD3 },
{ 0x00, 0xF4, 0xD4 },
{ 0x00, 0xF5, 0xD5 },
{ 0x00, 0xF6, 0xD6 },
{ 0x00, 0xF7, 0xF7 },
{ 0x00, 0xF8, 0xD8 },
{ 0x00, 0xF9, 0xD9 },
{ 0x00, 0xFA, 0xDA },
{ 0x00, 0xFB, 0xDB },
{ 0x00, 0xFC, 0xDC },
{ 0x00, 0xFD, 0xDD },
{ 0x00, 0xFE, 0xDE },
{ 0x00, 0xFF, 0xFF },
};


struct cs_info iso14_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x01, 0xa2, 0xa1 },
{ 0x00, 0xa2, 0xa1 },
{ 0x00, 0xa3, 0xa3 },
{ 0x01, 0xa5, 0xa4 },
{ 0x00, 0xa5, 0xa4 },
{ 0x01, 0xa6, 0xab },
{ 0x00, 0xa7, 0xa7 },
{ 0x01, 0xb8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x01, 0xba, 0xaa },
{ 0x00, 0xab, 0xa6 },
{ 0x01, 0xbc, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x01, 0xff, 0xaf },
{ 0x01, 0xb1, 0xb0 },
{ 0x00, 0xb1, 0xb0 },
{ 0x01, 0xb3, 0xb2 },
{ 0x00, 0xb3, 0xb2 },
{ 0x01, 0xb5, 0xb4 },
{ 0x00, 0xb5, 0xb4 },
{ 0x00, 0xb6, 0xb6 },
{ 0x01, 0xb9, 0xb7 },
{ 0x00, 0xb8, 0xa8 },
{ 0x00, 0xb9, 0xb6 },
{ 0x00, 0xba, 0xaa },
{ 0x01, 0xbf, 0xbb },
{ 0x00, 0xbc, 0xac },
{ 0x01, 0xbe, 0xbd },
{ 0x00, 0xbe, 0xbd },
{ 0x00, 0xbf, 0xbb },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x01, 0xf7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xd7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xff },
};

struct cs_info iso15_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x01, 0xa8, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa6 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x01, 0xb8, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb4 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x01, 0xbd, 0xbc },
{ 0x00, 0xbd, 0xbc },
{ 0x01, 0xff, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x01, 0xe0, 0xc0 },
{ 0x01, 0xe1, 0xc1 },
{ 0x01, 0xe2, 0xc2 },
{ 0x01, 0xe3, 0xc3 },
{ 0x01, 0xe4, 0xc4 },
{ 0x01, 0xe5, 0xc5 },
{ 0x01, 0xe6, 0xc6 },
{ 0x01, 0xe7, 0xc7 },
{ 0x01, 0xe8, 0xc8 },
{ 0x01, 0xe9, 0xc9 },
{ 0x01, 0xea, 0xca },
{ 0x01, 0xeb, 0xcb },
{ 0x01, 0xec, 0xcc },
{ 0x01, 0xed, 0xcd },
{ 0x01, 0xee, 0xce },
{ 0x01, 0xef, 0xcf },
{ 0x01, 0xf0, 0xd0 },
{ 0x01, 0xf1, 0xd1 },
{ 0x01, 0xf2, 0xd2 },
{ 0x01, 0xf3, 0xd3 },
{ 0x01, 0xf4, 0xd4 },
{ 0x01, 0xf5, 0xd5 },
{ 0x01, 0xf6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x01, 0xf8, 0xd8 },
{ 0x01, 0xf9, 0xd9 },
{ 0x01, 0xfa, 0xda },
{ 0x01, 0xfb, 0xdb },
{ 0x01, 0xfc, 0xdc },
{ 0x01, 0xfd, 0xdd },
{ 0x01, 0xfe, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xc0 },
{ 0x00, 0xe1, 0xc1 },
{ 0x00, 0xe2, 0xc2 },
{ 0x00, 0xe3, 0xc3 },
{ 0x00, 0xe4, 0xc4 },
{ 0x00, 0xe5, 0xc5 },
{ 0x00, 0xe6, 0xc6 },
{ 0x00, 0xe7, 0xc7 },
{ 0x00, 0xe8, 0xc8 },
{ 0x00, 0xe9, 0xc9 },
{ 0x00, 0xea, 0xca },
{ 0x00, 0xeb, 0xcb },
{ 0x00, 0xec, 0xcc },
{ 0x00, 0xed, 0xcd },
{ 0x00, 0xee, 0xce },
{ 0x00, 0xef, 0xcf },
{ 0x00, 0xf0, 0xd0 },
{ 0x00, 0xf1, 0xd1 },
{ 0x00, 0xf2, 0xd2 },
{ 0x00, 0xf3, 0xd3 },
{ 0x00, 0xf4, 0xd4 },
{ 0x00, 0xf5, 0xd5 },
{ 0x00, 0xf6, 0xd6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xd8 },
{ 0x00, 0xf9, 0xd9 },
{ 0x00, 0xfa, 0xda },
{ 0x00, 0xfb, 0xdb },
{ 0x00, 0xfc, 0xdc },
{ 0x00, 0xfd, 0xdd },
{ 0x00, 0xfe, 0xde },
{ 0x00, 0xff, 0xbe },
};

struct cs_info iscii_devanagari_tbl[] = {
{ 0x00, 0x00, 0x00 },
{ 0x00, 0x01, 0x01 },
{ 0x00, 0x02, 0x02 },
{ 0x00, 0x03, 0x03 },
{ 0x00, 0x04, 0x04 },
{ 0x00, 0x05, 0x05 },
{ 0x00, 0x06, 0x06 },
{ 0x00, 0x07, 0x07 },
{ 0x00, 0x08, 0x08 },
{ 0x00, 0x09, 0x09 },
{ 0x00, 0x0a, 0x0a },
{ 0x00, 0x0b, 0x0b },
{ 0x00, 0x0c, 0x0c },
{ 0x00, 0x0d, 0x0d },
{ 0x00, 0x0e, 0x0e },
{ 0x00, 0x0f, 0x0f },
{ 0x00, 0x10, 0x10 },
{ 0x00, 0x11, 0x11 },
{ 0x00, 0x12, 0x12 },
{ 0x00, 0x13, 0x13 },
{ 0x00, 0x14, 0x14 },
{ 0x00, 0x15, 0x15 },
{ 0x00, 0x16, 0x16 },
{ 0x00, 0x17, 0x17 },
{ 0x00, 0x18, 0x18 },
{ 0x00, 0x19, 0x19 },
{ 0x00, 0x1a, 0x1a },
{ 0x00, 0x1b, 0x1b },
{ 0x00, 0x1c, 0x1c },
{ 0x00, 0x1d, 0x1d },
{ 0x00, 0x1e, 0x1e },
{ 0x00, 0x1f, 0x1f },
{ 0x00, 0x20, 0x20 },
{ 0x00, 0x21, 0x21 },
{ 0x00, 0x22, 0x22 },
{ 0x00, 0x23, 0x23 },
{ 0x00, 0x24, 0x24 },
{ 0x00, 0x25, 0x25 },
{ 0x00, 0x26, 0x26 },
{ 0x00, 0x27, 0x27 },
{ 0x00, 0x28, 0x28 },
{ 0x00, 0x29, 0x29 },
{ 0x00, 0x2a, 0x2a },
{ 0x00, 0x2b, 0x2b },
{ 0x00, 0x2c, 0x2c },
{ 0x00, 0x2d, 0x2d },
{ 0x00, 0x2e, 0x2e },
{ 0x00, 0x2f, 0x2f },
{ 0x00, 0x30, 0x30 },
{ 0x00, 0x31, 0x31 },
{ 0x00, 0x32, 0x32 },
{ 0x00, 0x33, 0x33 },
{ 0x00, 0x34, 0x34 },
{ 0x00, 0x35, 0x35 },
{ 0x00, 0x36, 0x36 },
{ 0x00, 0x37, 0x37 },
{ 0x00, 0x38, 0x38 },
{ 0x00, 0x39, 0x39 },
{ 0x00, 0x3a, 0x3a },
{ 0x00, 0x3b, 0x3b },
{ 0x00, 0x3c, 0x3c },
{ 0x00, 0x3d, 0x3d },
{ 0x00, 0x3e, 0x3e },
{ 0x00, 0x3f, 0x3f },
{ 0x00, 0x40, 0x40 },
{ 0x01, 0x61, 0x41 },
{ 0x01, 0x62, 0x42 },
{ 0x01, 0x63, 0x43 },
{ 0x01, 0x64, 0x44 },
{ 0x01, 0x65, 0x45 },
{ 0x01, 0x66, 0x46 },
{ 0x01, 0x67, 0x47 },
{ 0x01, 0x68, 0x48 },
{ 0x01, 0x69, 0x49 },
{ 0x01, 0x6a, 0x4a },
{ 0x01, 0x6b, 0x4b },
{ 0x01, 0x6c, 0x4c },
{ 0x01, 0x6d, 0x4d },
{ 0x01, 0x6e, 0x4e },
{ 0x01, 0x6f, 0x4f },
{ 0x01, 0x70, 0x50 },
{ 0x01, 0x71, 0x51 },
{ 0x01, 0x72, 0x52 },
{ 0x01, 0x73, 0x53 },
{ 0x01, 0x74, 0x54 },
{ 0x01, 0x75, 0x55 },
{ 0x01, 0x76, 0x56 },
{ 0x01, 0x77, 0x57 },
{ 0x01, 0x78, 0x58 },
{ 0x01, 0x79, 0x59 },
{ 0x01, 0x7a, 0x5a },
{ 0x00, 0x5b, 0x5b },
{ 0x00, 0x5c, 0x5c },
{ 0x00, 0x5d, 0x5d },
{ 0x00, 0x5e, 0x5e },
{ 0x00, 0x5f, 0x5f },
{ 0x00, 0x60, 0x60 },
{ 0x00, 0x61, 0x41 },
{ 0x00, 0x62, 0x42 },
{ 0x00, 0x63, 0x43 },
{ 0x00, 0x64, 0x44 },
{ 0x00, 0x65, 0x45 },
{ 0x00, 0x66, 0x46 },
{ 0x00, 0x67, 0x47 },
{ 0x00, 0x68, 0x48 },
{ 0x00, 0x69, 0x49 },
{ 0x00, 0x6a, 0x4a },
{ 0x00, 0x6b, 0x4b },
{ 0x00, 0x6c, 0x4c },
{ 0x00, 0x6d, 0x4d },
{ 0x00, 0x6e, 0x4e },
{ 0x00, 0x6f, 0x4f },
{ 0x00, 0x70, 0x50 },
{ 0x00, 0x71, 0x51 },
{ 0x00, 0x72, 0x52 },
{ 0x00, 0x73, 0x53 },
{ 0x00, 0x74, 0x54 },
{ 0x00, 0x75, 0x55 },
{ 0x00, 0x76, 0x56 },
{ 0x00, 0x77, 0x57 },
{ 0x00, 0x78, 0x58 },
{ 0x00, 0x79, 0x59 },
{ 0x00, 0x7a, 0x5a },
{ 0x00, 0x7b, 0x7b },
{ 0x00, 0x7c, 0x7c },
{ 0x00, 0x7d, 0x7d },
{ 0x00, 0x7e, 0x7e },
{ 0x00, 0x7f, 0x7f },
{ 0x00, 0x80, 0x80 },
{ 0x00, 0x81, 0x81 },
{ 0x00, 0x82, 0x82 },
{ 0x00, 0x83, 0x83 },
{ 0x00, 0x84, 0x84 },
{ 0x00, 0x85, 0x85 },
{ 0x00, 0x86, 0x86 },
{ 0x00, 0x87, 0x87 },
{ 0x00, 0x88, 0x88 },
{ 0x00, 0x89, 0x89 },
{ 0x00, 0x8a, 0x8a },
{ 0x00, 0x8b, 0x8b },
{ 0x00, 0x8c, 0x8c },
{ 0x00, 0x8d, 0x8d },
{ 0x00, 0x8e, 0x8e },
{ 0x00, 0x8f, 0x8f },
{ 0x00, 0x90, 0x90 },
{ 0x00, 0x91, 0x91 },
{ 0x00, 0x92, 0x92 },
{ 0x00, 0x93, 0x93 },
{ 0x00, 0x94, 0x94 },
{ 0x00, 0x95, 0x95 },
{ 0x00, 0x96, 0x96 },
{ 0x00, 0x97, 0x97 },
{ 0x00, 0x98, 0x98 },
{ 0x00, 0x99, 0x99 },
{ 0x00, 0x9a, 0x9a },
{ 0x00, 0x9b, 0x9b },
{ 0x00, 0x9c, 0x9c },
{ 0x00, 0x9d, 0x9d },
{ 0x00, 0x9e, 0x9e },
{ 0x00, 0x9f, 0x9f },
{ 0x00, 0xa0, 0xa0 },
{ 0x00, 0xa1, 0xa1 },
{ 0x00, 0xa2, 0xa2 },
{ 0x00, 0xa3, 0xa3 },
{ 0x00, 0xa4, 0xa4 },
{ 0x00, 0xa5, 0xa5 },
{ 0x00, 0xa6, 0xa6 },
{ 0x00, 0xa7, 0xa7 },
{ 0x00, 0xa8, 0xa8 },
{ 0x00, 0xa9, 0xa9 },
{ 0x00, 0xaa, 0xaa },
{ 0x00, 0xab, 0xab },
{ 0x00, 0xac, 0xac },
{ 0x00, 0xad, 0xad },
{ 0x00, 0xae, 0xae },
{ 0x00, 0xaf, 0xaf },
{ 0x00, 0xb0, 0xb0 },
{ 0x00, 0xb1, 0xb1 },
{ 0x00, 0xb2, 0xb2 },
{ 0x00, 0xb3, 0xb3 },
{ 0x00, 0xb4, 0xb4 },
{ 0x00, 0xb5, 0xb5 },
{ 0x00, 0xb6, 0xb6 },
{ 0x00, 0xb7, 0xb7 },
{ 0x00, 0xb8, 0xb8 },
{ 0x00, 0xb9, 0xb9 },
{ 0x00, 0xba, 0xba },
{ 0x00, 0xbb, 0xbb },
{ 0x00, 0xbc, 0xbc },
{ 0x00, 0xbd, 0xbd },
{ 0x00, 0xbe, 0xbe },
{ 0x00, 0xbf, 0xbf },
{ 0x00, 0xc0, 0xc0 },
{ 0x00, 0xc1, 0xc1 },
{ 0x00, 0xc2, 0xc2 },
{ 0x00, 0xc3, 0xc3 },
{ 0x00, 0xc4, 0xc4 },
{ 0x00, 0xc5, 0xc5 },
{ 0x00, 0xc6, 0xc6 },
{ 0x00, 0xc7, 0xc7 },
{ 0x00, 0xc8, 0xc8 },
{ 0x00, 0xc9, 0xc9 },
{ 0x00, 0xca, 0xca },
{ 0x00, 0xcb, 0xcb },
{ 0x00, 0xcc, 0xcc },
{ 0x00, 0xcd, 0xcd },
{ 0x00, 0xce, 0xce },
{ 0x00, 0xcf, 0xcf },
{ 0x00, 0xd0, 0xd0 },
{ 0x00, 0xd1, 0xd1 },
{ 0x00, 0xd2, 0xd2 },
{ 0x00, 0xd3, 0xd3 },
{ 0x00, 0xd4, 0xd4 },
{ 0x00, 0xd5, 0xd5 },
{ 0x00, 0xd6, 0xd6 },
{ 0x00, 0xd7, 0xd7 },
{ 0x00, 0xd8, 0xd8 },
{ 0x00, 0xd9, 0xd9 },
{ 0x00, 0xda, 0xda },
{ 0x00, 0xdb, 0xdb },
{ 0x00, 0xdc, 0xdc },
{ 0x00, 0xdd, 0xdd },
{ 0x00, 0xde, 0xde },
{ 0x00, 0xdf, 0xdf },
{ 0x00, 0xe0, 0xe0 },
{ 0x00, 0xe1, 0xe1 },
{ 0x00, 0xe2, 0xe2 },
{ 0x00, 0xe3, 0xe3 },
{ 0x00, 0xe4, 0xe4 },
{ 0x00, 0xe5, 0xe5 },
{ 0x00, 0xe6, 0xe6 },
{ 0x00, 0xe7, 0xe7 },
{ 0x00, 0xe8, 0xe8 },
{ 0x00, 0xe9, 0xe9 },
{ 0x00, 0xea, 0xea },
{ 0x00, 0xeb, 0xeb },
{ 0x00, 0xec, 0xec },
{ 0x00, 0xed, 0xed },
{ 0x00, 0xee, 0xee },
{ 0x00, 0xef, 0xef },
{ 0x00, 0xf0, 0xf0 },
{ 0x00, 0xf1, 0xf1 },
{ 0x00, 0xf2, 0xf2 },
{ 0x00, 0xf3, 0xf3 },
{ 0x00, 0xf4, 0xf4 },
{ 0x00, 0xf5, 0xf5 },
{ 0x00, 0xf6, 0xf6 },
{ 0x00, 0xf7, 0xf7 },
{ 0x00, 0xf8, 0xf8 },
{ 0x00, 0xf9, 0xf9 },
{ 0x00, 0xfa, 0xfa },
{ 0x00, 0xfb, 0xfb },
{ 0x00, 0xfc, 0xfc },
{ 0x00, 0xfd, 0xfd },
{ 0x00, 0xfe, 0xfe },
{ 0x00, 0xff, 0xff },
};

struct enc_entry encds[] = {
{"ISO8859-1",iso1_tbl},
{"ISO8859-2",iso2_tbl},
{"ISO8859-3",iso3_tbl},
{"ISO8859-4",iso4_tbl},
{"ISO8859-5",iso5_tbl},
{"ISO8859-6",iso6_tbl},
{"ISO8859-7",iso7_tbl},
{"ISO8859-8",iso8_tbl},
{"ISO8859-9",iso9_tbl},
{"ISO8859-10",iso10_tbl},
{"KOI8-R",koi8r_tbl},
{"KOI8-U",koi8u_tbl},
{"microsoft-cp1251",cp1251_tbl},
{"ISO8859-13", iso13_tbl},
{"ISO8859-14", iso14_tbl},
{"ISO8859-15", iso15_tbl},
{"ISCII-DEVANAGARI", iscii_devanagari_tbl},
};


struct cs_info * get_current_cs(const char * es) {
  struct cs_info * ccs = encds[0].cs_table;
  int n = sizeof(encds) / sizeof(encds[0]);
  for (int i = 0; i < n; i++) {
    if (strcmp(es,encds[i].enc_name) == 0) {
      ccs = encds[i].cs_table;
    }
  }
  return ccs;
}

struct unicode_info * get_utf_cs() {
  return utf_lst;
}

int get_utf_cs_len() {
  return UTF_LST_LEN;
}

struct lang_map lang2enc[] = {
{"az", "UTF-8", LANG_az},
{"bg", "microsoft-cp1251", LANG_bg},
{"ca", "ISO8859-1", LANG_ca},
{"cs", "ISO8859-2", LANG_cs},
{"da", "ISO8859-1", LANG_da},
{"de", "ISO8859-1", LANG_de},
{"el", "ISO8859-7", LANG_el},
{"en", "ISO8859-1", LANG_en},
{"es", "ISO8859-1", LANG_es},
{"eu", "ISO8859-1", LANG_eu},
{"gl", "ISO8859-1", LANG_gl},
{"fr", "ISO8859-15", LANG_fr},
{"hr", "ISO8859-2", LANG_hr},
{"hu", "ISO8859-2", LANG_hu},
{"it", "ISO8859-1", LANG_it},
{"la", "ISO8859-1", LANG_la},
{"lv", "ISO8859-13", LANG_lv},
{"nl", "ISO8859-1", LANG_nl},
{"pl", "ISO8859-2", LANG_pl},
{"pt", "ISO8859-1", LANG_pt},
{"sv", "ISO8859-1", LANG_sv},
{"tr", "UTF-8", LANG_tr},
{"ru", "KOI8-R", LANG_ru},
{"uk", "KOI8-U", LANG_uk}
};


const char * get_default_enc(const char * lang) {
  int n = sizeof(lang2enc) / sizeof(lang2enc[0]);
  for (int i = 0; i < n; i++) {
    if (strcmp(lang,lang2enc[i].lang) == 0) {
      return lang2enc[i].def_enc;
    }
  }
  return NULL;
}

int get_lang_num(const char * lang) {
  int n = sizeof(lang2enc) / sizeof(lang2enc[0]);
  for (int i = 0; i < n; i++) {
    if (strncmp(lang,lang2enc[i].lang,2) == 0) {
      return lang2enc[i].num;
    }
  }
  return LANG_xx;
}
